const
  CMDPORT    = $50;
  DATAPORT   = $51;

  cmdRESET    = $00;
  cmdASCII    = $01;
  cmdSHASCII  = $02;
  cmdSTSR     = $03;
  cmdSTORNO   = $04;
  cmdREV      = $05;
  cmdREVD     = $06;
  cmdFDDMOUNT = $10;
  cmdGETFREE  = $20;
  cmdCHDIR    = $21;
  cmdGETCWD   = $22;
  cmdSTAT     = $30;
  cmdUNLNIK   = $31;
  cmdCHMOD    = $32;
  cmdUTIME    = $33;
  cmdRENAME   = $34;
  cmdMKDIR    = $40;
  cmdREADDIR  = $41;
  cmdFILELIST = $42;
  cmdNEXT     = $43;
  cmdOPEN     = $50;
  cmdSEEK     = $51;
  cmdTRUNC    = $52;
  cmdSYNC     = $53;
  cmdCLOSE    = $54;
  cmdTELL     = $55;
  cmdSIZE     = $56;
  cmdRTCSETD  = $60;
  cmdRTCSETT  = $61;
  cmdRTCGETD  = $62;
  cmdRTCGETT  = $63;

  FA_OPEN_EXISTING = $00;
  FA_READ          = $01;
  FA_WRITE         = $02;
  FA_CREATE_NEW    = $04;
  FA_CREATE_ALWAYS = $08;
  FA_OPEN_ALWAYS   = $10;
  FILE_SHASCII_CNV = $20;

type
  HexStr = string[4];
  Str16 = string[16];
  Str254 = string[254];

var
  UCRev : integer;
  UCBuf : array [0..255] of byte;
  UCStatMaster : byte;
  UCStatLastCmd : byte;
  UCStatCode : byte;
  UCStatFatFS : byte;
  AskFnm, AskFnm2 : Str254;


procedure Help;
begin
  WriteLn('UNIUTIL /GETRTCTIME             - show UniCard RTC time');
  WriteLn('UNIUTIL /GRTR                   - show Unicard RTC time while not key pressed');
  WriteLn('UNIUTIL /SETRTCTIME HH:MM:SS    - set UniCard RTC time to HH:MM:SS');
  WriteLn('UNIUTIL /SETCPMTIME             - set CP/M time to UniCard RTC time');
  WriteLn('UNIUTIL /SAVERD page(0-1) bank(0-16) size(0-16) [filename] - RD to SD');
  WriteLn('UNIUTIL /FILETOSD [filename]    - copy file from CP/M to root of SD card');
  WriteLn('UNIUTIL /FILEFROMSD [filename]  - copy file from root of SD card to CP/M');
end;

function HexFmt(i:integer;w:integer):HexStr;
begin
  { if (w<=1) and (i and $FFF0 = 0) then
    HexFmt := Hex[i]
  else
    HexFmt := HexFmt(i shr 4, w-1) + Hex[i and $000F]; }
end;

function HexByte(i:integer):HexStr;
const Hex : array [0..$F] of char = '0123456789ABCDEF';
begin
  i := i and $00FF;
  HexByte := Hex[i shr 4] + Hex[i and $000F];
end;

function DecStr2(i:integer):Str16;
var r : Str16;
begin
  Str(i,r);
  if length(r)<2 then
    r := '0'+r;
  DecStr2 := r;
end;

procedure WriteTime(hh,mm,ss:integer);
begin
  Write(DecStr2(hh),':',DecStr2(mm),':',DecStr2(ss));
end;

procedure ReadUCStatus;
begin
  UCStatMaster := Port[CMDPORT];
  UCStatLastCmd := Port[CMDPORT];
  UCStatCode := Port[CMDPORT];
  UCStatFatFS := Port[CMDPORT];
end;

procedure WriteUCStatus;
begin
  WriteLn('UC status: Master=',HexByte(UCStatMaster),' LastCmd=',HexByte(UCStatLastCmd),
          ' Code=',HexByte(UCStatCode),' FatFS=',HexByte(UCStatFatFS));
end;

procedure WriteUCBuf(l:integer);
var i,j : byte;
begin
  WriteLn('UniCard buffer');
  for i:=0 to 15 do
    begin
      for j:=0 to 15 do
        if i*16+j<l then
          begin
            Write(HexByte(UCBuf[i*16+j]));
            Write(' ');
          end;
      if i*16+j<l then
        WriteLn;
    end;
end;

procedure ReadUCData(cnt : integer);
var pos : integer;
begin
  pos := 0;
  while pos<cnt do
    begin
      UCBuf[pos] := Port[DATAPORT];
      { Write('ReadData=',UCBuf[pos]); }
      pos := pos+1;
    end;
end;

procedure CheckUCRev;
var r : array [0..3] of byte;
begin
  Port[CMDPORT] := cmdRESET;
  Port[CMDPORT] := cmdREVD;
  ReadUCStatus;
  { WriteUCStatus; }
  if (UCStatMaster=2) and (UCStatLastCmd=6) and (UCStatCode=4) then
    begin
      ReadUCData(4);
      { WriteUCBuf(4); }
      if (UCBuf[2]=0) and (UCBuf[3]=0) then
        UCRev := UCBuf[1]*256 + UCBuf[0]
      else
        UCRev := 0;
    end
  else
    UCRev := 0;
end;

procedure GetRTCTime;
begin
  Port[CMDPORT] := cmdRTCGETT;
  ReadUCStatus;
  { WriteUCStatus; }
  if (UCStatMaster=2) and (UCStatLastCmd=99) and (UCStatCode=3) then
    begin
      ReadUCData(3);
      { WriteUCBuf(3); }
      Write('UniCard RTC time is ');
      WriteTime(UCBuf[0],UCBuf[1],UCBuf[2]);
      WriteLn;
    end
end;

procedure SetRTCTime(tstr:Str16);
var hh,mm,ss,ch,tch : integer;
begin
  if (length(tstr)<>8) or (tstr[3]<>':') or (tstr[6]<>':') then
    WriteLn('Wrong time format, use HH:MM:SS')
  else
    begin
      Val(Copy(tstr,1,2),hh,ch);
      tch := ch;
      Val(Copy(tstr,4,2),mm,ch);
      tch := tch+ch;
      Val(Copy(tstr,7,2),ss,ch);
      tch := tch+ch;
      if (tch=0) and (hh>=0) and (hh<24) and (mm>=00) and (mm<60) and (ss>=0) and (ss<60) then
        begin
          Write('Setting UniCard RTC time to ');
          WriteTime(hh,mm,ss);
          Write(' ... ');
          Port[CMDPORT] := cmdRTCSETT;
          ReadUCStatus;
          { WriteUCStatus; }
          if (UCStatMaster=1) and (UCStatLastCmd=97) and (UCStatCode=255) then
            begin
              Port[DATAPORT] := hh;
              Port[DATAPORT] := mm;
              Port[DATAPORT] := ss;
              ReadUCStatus;
              { WriteUCStatus; }
              if UCStatMaster=0 then
                WriteLn('OK')
              else
                WriteLn('error writing time to RTC');
            end;
          GetRTCTime;
        end
      else
        WriteLn('Wrong time format - no valid numbers, use HH:MM:SS');
    end;
end;

procedure SetCPMTime;
var f : Real;
    lb, mb, hb : Byte;
begin
  Port[CMDPORT] := cmdRTCGETT;
  ReadUCStatus;
  { WriteUCStatus; }
  if (UCStatMaster=2) and (UCStatLastCmd=99) and (UCStatCode=3) then
    begin
      ReadUCData(3);
      { WriteUCBuf(3); }
      Write('Setting CPM time to ');
      WriteTime(UCBuf[0],UCBuf[1],UCBuf[2]);
      WriteLn;
      { WriteLn('LB=',Mem[$EA40],' MB=',Mem[$EA41],' HB=',Mem[$EA42]); }
      f := UCBuf[0]*3600.0 + UCBuf[1]*60.0 + UCBuf[2];
      { WriteLn('F=',f); }
      hb := Trunc(f/65536.0);
      f := Frac(f/65536.0)*65536.0;
      mb := Trunc(f/256.0);
      f := Frac(f/256.0)*256.0;
      lb := Round(f);
      { WriteLn('LB=',lb,' MB=',mb,' HB=',hb); }
      Mem[$EA40] := lb;
      Mem[$EA41] := mb;
      Mem[$EA42] := hb;
    end
end;

function ReadKey:Char;
begin
  ReadKey := Chr(BIOS(2));
end;

procedure GetRTCTimeRepeat;
var key : char;
begin
  WriteLn('Showing Unicard RTC time. Press any key several times to abort.');
  while not KeyPressed do
    begin
      Port[CMDPORT] := cmdRTCGETT;
      ReadUCStatus;
      { WriteUCStatus; }
      if (UCStatMaster=2) and (UCStatLastCmd=99) and (UCStatCode=3) then
        begin
          ReadUCData(3);
          { WriteUCBuf(3); }
          WriteTime(UCBuf[0],UCBuf[1],UCBuf[2]);
          Write(' ');
          Delay(333);
          Write(#13);
        end;
    end;
end;

procedure SendTXParam(s:Str254);
var i:Integer;
begin
  { Write('TxtOut='); }
  for i:=1 to Length(s) do
    begin
      Port[DATAPORT]:=Ord(s[i]);
      { Write(HexByte(Ord(s[i]))); }
    end;
  WriteLn;
end;

procedure SaveRD(pg,bnk,sz:Str16;fnm:Str254);
var i,j,p,b,bb,s,ch,tch : Integer;
begin
  Val(pg,p,ch);
  tch:=ch;
  Val(bnk,b,ch);
  tch:=tch+ch;
  Val(sz,s,ch);
  tch:=tch+ch;
  if (tch=0) and (p>=0) and (p<=1) and (b>=0) and (b<=15) and (s>0) and (s<=16-b) then
    begin
      WriteLn('Writing ',s*64,' kB from RD to SD card file ''',fnm,'''');
      Port[CMDPORT]:=cmdRESET;
      Port[CMDPORT]:=cmdOPEN;
      Port[DATAPORT]:=FA_CREATE_ALWAYS+FA_READ+FA_WRITE;
      SendTXParam(fnm+#00);
      ReadUCStatus;
      { WriteUCStatus; }
      if (UCStatMaster=$38) and (UCStatLastCmd=cmdOPEN) and (UCStatCode=0) then
        begin
          Port[$E8]:=p;
          for bb:=b to b+s-1 do
            begin
              Port[$E9]:=bb;
              Port[$00EB] := $00;
              for j:=0 to 255 do
                begin
                  Write('Writing page ',p,', bank ',bb,', sector ',j,'   ');
                  for i:=0 to 255 do
                    begin
                      UCBuf[i] := Port[$EA];
                      Port[DATAPORT]:= UCBuf[i];
                    end;
                  { WriteUCBuf(256); }
                  Write(#13);
                end;
            end;
          Port[$E8]:=0;
          Port[$E9]:=0;
          { WriteUCBuf(256); }
          Port[CMDPORT]:=cmdCLOSE;
          ReadUCStatus;
          { WriteUCStatus; }
          if (UCStatMaster=0) and (UCStatLastCmd=cmdCLOSE) and (UCStatCode=0) then
            WriteLn('File ''',fnm,''' created successfuly.              ')
          else
            WriteLn('Cannot close file ''',fnm,''', FS error ',UCStatFatFS);
        end
      else
        WriteLn('Cannot open file ''',fnm,''', FS error ',UCStatFatFS);
    end
  else
    WriteLn('Bad page, bank or size parameters.');
end;

procedure FileToSD(cpmfnm,sdfnm:Str254);
var i,fs : Integer;
    f : file of Byte;
    tmpb : Byte;
begin
  WriteLn('Assigning file ',cpmfnm);
  Assign(f,cpmfnm);
  {$i-} Reset(f); {$i+}
  Write('Vysledek: ');
  WriteLn('VYSLEDEK=',ioresult);
  if ioresult=0 then
    begin
      WriteLn('Getting filesize');
      fs := FileSize(f);
      WriteLn('Copy ',fs,' B from CP/M file ',cpmfnm,' to SD file ',sdfnm);
      Port[CMDPORT]:=cmdRESET;
      Port[CMDPORT]:=cmdOPEN;
      Port[DATAPORT]:=FA_CREATE_ALWAYS+FA_READ+FA_WRITE;
      SendTXParam(sdfnm+#00);
      ReadUCStatus;
      WriteUCStatus;
      if (UCStatMaster=$38) and (UCStatLastCmd=cmdOPEN) and (UCStatCode=0) then
        begin
          i:=0;
          while not Eof(f) do
            begin
              Write('Writing byte ',i,'    ');
              i:=i+1;
              Read(f,tmpb);
              Write('to SD     ');
              Port[DATAPORT]:=tmpb;
              Write(#13)
            end;
          Port[CMDPORT]:=cmdCLOSE;
          ReadUCStatus;
          WriteUCStatus;
          if (UCStatMaster=0) and (UCStatLastCmd=cmdCLOSE) and (UCStatCode=0) then
            WriteLn('SD file ''',sdfnm,''' created successfuly.              ')
          else
            WriteLn('Cannot close SD file ''',sdfnm,''', FS error ',UCStatFatFS);
        end
      else
        WriteLn('Cannot open SD file ''',sdfnm,''', FS error ',UCStatFatFS);
    end
  else
    WriteLn('Cannot open CP/M file ',cpmfnm,', error ',ioresult);
end;

begin
  Write('Unicard Utilities 1.0a ');
  CheckUCRev;
  if UCRev>0 then
    Write('(Unicard OK, fw.',UCRev,')')
  else
    Write('(no Unicard detected)');
  WriteLn('          (c) 2011 Martin Lukasek');
  WriteLn;
  if (1=1) or (ParamCount>0) then
    begin
      if UCRev>=37 then
        begin
          if (ParamStr(1)='/GETRTCTIME') or (ParamStr(1)='/GRT') then
            GetRTCTime
          else
          if (ParamStr(1)='/SETRTCTIME') or (ParamStr(1)='/SRT') then
            SetRTCTime(ParamStr(2))
          else
          if (ParamStr(1)='/SETCPMTIME') or (ParamStr(1)='/SCT') then
            SetCPMTime
          else
          if (1=0) or (ParamStr(1)='/GRTR') then
            GetRTCTimeRepeat
          else
          if (1=0) or (ParamStr(1)='/SAVERD') then
            begin
              if (ParamCount<5) or (ParamStr(5)='ASK') then
                begin
                  Write('Filename (full path): ');
                  ReadLn(AskFnm);
                end
              else
                AskFnm:=ParamStr(5);
              SaveRD(ParamStr(2),ParamStr(3),ParamStr(4),AskFnm)
            end
          else
          if (1=0) or (ParamStr(1)='/FILETOSD') or (ParamStr(1)='/FTS') then
            begin
              if (ParamCount=1) or (ParamStr(2)='ASK') then
                begin
                  Write('Source filename (CP/M path): ');
                  ReadLn(AskFnm);
                  Write('Target filename (SD card full path): ');
                  ReadLn(AskFnm2);
                end
              else
                begin
                  AskFnm:=ParamStr(2); { 'UU.BAK'};
                  AskFnm2:='/'+AskFnm;
                end;
              FileToSD(AskFnm,AskFnm2);
            end
          else
            Help;
        end
      else
        WriteLn('Low Unicard firmware version, please update you firmware.');
    end
  else
    Help;
end.
